# -*- coding: utf-8 -*-
#!/usr/bin/env python3

# Authorship :
#   Robert VISEUR <robert.viseur@umons.ac.be>
# License : GPL version 3

import warnings

import sys
import os.path
import os
import configparser

import socket
import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse
from urllib.parse import urljoin

import sqlite3

from datetime import datetime
import time

import re

# test the internet connection
def isActiveInternetConnection():
    bConnection = True
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(("8.8.8.8", 80))
    except Exception:
        bConnection = False
    finally:
        sock.close()
    return bConnection

# extract the full domain from URL
def getFullDomainFromUrl (url):
    o1 = urlparse(url)
    return o1[1]

# extract the domain name from full domain
def getDomainNameFromFullDomain(domain):
    tlds = [".com", ".net", ".org",
            ".biz", ".info", ".name", ".coop",
            ".edu", ".gov", ".mil",
            ".eu", ".be", ".ch", ".co.uk", ".de", ".es", ".fr", ".nl", ".no", ".ru", ".se",
            ".au", ".br", ".ca", ".cn", ".co", ".in", ".jp", ".us"]
    tmp = domain
    if tmp=="":
        md = ""
    else:
        md = ""
        
        tmpa = tmp.split(".")
        for i in range (0,len(tmpa)+1):
            last = tmpa[i:]
            tldc = ".".join(last)
            if "."+tldc in tlds:
                md = ".".join(tmpa[i-1:])
                break
    return md

# extract the domain name for url
def getDomainNameFromUrl(url):
    out = urlparse(url)
    return getDomainNameFromFullDomain(out[1])

# get stealth headers
def getStealthHeaders ():
    headers = {
        'User-Agent':'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.97 Safari/537.36',
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Language': 'en-US,en;q=0.5',
        'Connection': 'keep-alive',
        'Upgrade-Insecure-Requests': '1',
        'Pragma': 'no-cache',
        'Cache-Control': 'no-cache'
    }
    return headers

# detect non-HTML File (Url)
def isNotHtmlFile (url):
    bTest = False
    exts = [".pdf",".doc",".docx",".jpg",".png",".mp3",".avi",".mp4", ".exe", ".zip"]
    for i in range (0, len(exts)):
        tmp = url.lower()
        if tmp.endswith(exts[i]):
            bTest = True
    schms = ["mailto:", "phone:", "tel:", "javascript:"]
    for i in range (0, len(schms)):
        tmp = url.lower()
        if re.search(schms[i],tmp) is not None:
            bTest = True
    return bTest

# load URLs from file(for crawl)
def loadUrlsFromFile (filename):
    with open(filename, "r") as filin:
        tmp = filin.read().strip()
    urls = tmp.split("\n")
    return urls

# create the database
def createDatabase(filename):
    con = sqlite3.connect(filename)
    cur = con.cursor()
    cur.execute("CREATE TABLE IF NOT EXISTS urls_init (url TEXT NOT NULL, fulldomain TEXT NOT NULL, domainname TEXT NOT NULL, UNIQUE(url));")
    cur.execute("CREATE TABLE IF NOT EXISTS links (hyperlink_from TEXT NOT NULL, hyperlink_to TEXT NOT NULL, fulldomain_from TEXT NOT NULL, fulldomain_to TEXT NOT NULL, domainname_from TEXT NOT NULL, domainname_to TEXT NOT NULL);")
    cur.execute("CREATE TABLE IF NOT EXISTS urls (url TEXT NOT NULL, fulldomain TEXT NOT NULL, domainname TEXT NOT NULL, iscrawled BOOLEAN, iserror BOOLEAN, isignored BOOLEAN, UNIQUE(url));")
    con.commit()
    con.close()
    return True

# add the first URLs to the database
def addFirstUrlsToDatabase(filename, urls):
    con = sqlite3.connect(filename)
    cur = con.cursor()
    for i in range (0,len(urls)):
        try:
            cur.execute("INSERT INTO urls_init (url, fulldomain, domainname) VALUES ('%s', '%s', '%s');" % (urls[i], getDomainNameFromUrl(urls[i]), getDomainNameFromUrl(urls[i])))
        except:
            pass
    con.commit()
    con.close()
    return True

# add the extracted URLs to database
def addUrlsToDatabase(filename, urls):
    con = sqlite3.connect(filename)
    cur = con.cursor()
    for i in range (0,len(urls)):
        try:
            cur.execute("INSERT INTO urls (url, fulldomain, domainname, iscrawled, iserror, isignored) VALUES ('%s', '%s', '%s', 0, 0, 0);" % (urls[i], getFullDomainFromUrl(urls[i]), getDomainNameFromUrl(urls[i])))
        except:
            pass
    con.commit()
    con.close()
    return True

# add the hyperlinks to database
def addLinksToDatabase(filename, hyperlink_from, hyperlink_to):
    con = sqlite3.connect(filename)
    cur = con.cursor()
    for i in range (0,len(hyperlink_from)):
        try:
            cur.execute("INSERT INTO links (hyperlink_from, hyperlink_to, fulldomain_from, fulldomain_to, domainname_from, domainname_to) VALUES ('%s', '%s', '%s', '%s', '%s', '%s');" % (hyperlink_from[i], hyperlink_to[i], getFullDomainFromUrl(hyperlink_from[i]), getFullDomainFromUrl(hyperlink_to[i]), getDomainNameFromUrl(hyperlink_from[i]), getDomainNameFromUrl(hyperlink_to[i])))
        except:
            pass
    con.commit()
    con.close()
    return True

# update an URL to database
def updateUrlsInDatabase(filename, url, what, value):
    con = sqlite3.connect(filename)
    cur = con.cursor()
    try:
        cur.execute("UPDATE urls SET %s = %d WHERE url = '%s';" % (what, value, url))
    except:
        pass
    con.commit()
    con.close()
    return True

# read the argument(s)
# first = the Python filename
parameter_1 = None
parameter_2 = None
parameter_3 = None
parameter_4 = None
arg = []
for argtmp in sys.argv:
    arg.append(argtmp)
cli_cl = []
if (len(arg)>1):
    for i in range (1, len(arg)):
        if i==1:
            parameter_1 = arg[i]
        if i==2:
            parameter_2 = arg[i]
        if i==3:
            parameter_3 = arg[i]
        if i==4:
            parameter_4 = arg[i]

config = configparser.ConfigParser()
config.read("run.cfg")
#exit()

if (parameter_1=="init"):
    print ("Initiating the crawl...")
    if parameter_2 is not None:
        if not os.path.exists(parameter_2):
            os.mkdir(parameter_2)
        filename_urls = parameter_2 + ".txt"
        filename_dbs  = os.path.join(parameter_2, parameter_2 + ".db")
        if (os.path.isfile(filename_urls)):
            print ("  The file with URLs exists.")
        else:
            print ("  The file with URLs doesn't exist.")
        if os.path.isfile(filename_urls):
            tobecrawled = loadUrlsFromFile (filename_urls)
            createDatabase(filename_dbs)
            addFirstUrlsToDatabase (filename_dbs, tobecrawled)
            print ("  The file of database was created.")
    else:
        print ("  A parameter is missing.")  
    print ("End.")
elif  (parameter_1=="check"):
    print ("Checking the files...")
    if parameter_2 is not None:
        filename_urls = os.path.join(parameter_2, parameter_2 + ".txt")
        filename_dbs  = os.path.join(parameter_2, parameter_2 + ".db")
        if (os.path.isfile(filename_urls)):
            print ("  The file with URLs exists.")
        else:
            print ("  The file with URLs doesn't exist.")
        if (os.path.isfile(filename_dbs)):
            print ("  The file of database exists.")
        else:
            print ("  The file of database doesn't exist.")
    else:
        print ("  A parameter is missing.")    
    print ("End.")
elif  (parameter_1=="crawl"):

    if parameter_2 is not None and parameter_3 is not None:

        filename_urls = parameter_3 + ".txt"
        filename_dbs  = os.path.join(parameter_3, parameter_3 + ".db")
        go = True
        
        if (os.path.isfile(filename_urls)):
            print ("  The file with URLs exists.")
        else:
            print ("  The file with URLs doesn't exist.")
            go = False
        if (os.path.isfile(filename_dbs)):
            print ("  The file of database exists.")
        else:
            print ("  The file of database doesn't exist.")
            go = False

        limit = int(config['DEFAULT']['limit'])
        stealth = int(config['DEFAULT']['stealth'])
        timeout = int(config['DEFAULT']['timeout'])
        depth = int(config['DEFAULT']['depth'])
        
        if (go==True):
            start = time.time()

            print ("Crawling the files...")

            firstdomains   = []
            wascrawled     = []
            allurls        = []
            hyperlink_from = []
            hyperlink_to   = []

            tobecrawled = loadUrlsFromFile (filename_urls)
            allurls     = loadUrlsFromFile (filename_urls) 
            for i in range (0,len(tobecrawled)):
                firstdomains.append(getFullDomainFromUrl(tobecrawled[i]))

            # cofiguration of the user agent
            if stealth==1:
                headers = getStealthHeaders()
            else:
                headers = {}

            print ("Starting the crawl (from %d links)..." % len(tobecrawled))

            i = 0
            depth_n = []
            n = len(tobecrawled)
            depth_n.append(n)
            depth_i = 1
            while (i < n) and (i < limit) and (depth_i <= depth):
                url = tobecrawled[i]
                print ("  Starting %d : %s (%d, %d)." % (i+1, url, int(getFullDomainFromUrl(url) in firstdomains), int(not isNotHtmlFile(url))))

                if i==0:
                    addUrlsToDatabase(filename_dbs, tobecrawled)
                
                # read the page (2 modes : inside and outside)
                if url not in wascrawled and (parameter_2=="outside" or (parameter_2=="inside" and getFullDomainFromUrl(url) in firstdomains)):
                    
                    try:
                        if not isNotHtmlFile(url):
                            while not(isActiveInternetConnection()):
                                print("  A network failure was detected...")
                                time.sleep(10)
                            try:
                                reqs = requests.get(url, headers=headers,timeout=timeout)
                            except (requests.exceptions.SSLError):
                                # crappy handling of ssl error
                                warnings.simplefilter("ignore")
                                reqs = requests.get(url, headers=headers,timeout=timeout, verify = False)
                                warnings.resetwarnings()
                                print("  An SSL error was bypassed.")

                            wascrawled.append(url)
                            updateUrlsInDatabase(filename_dbs, url, "iscrawled", 1)
                        else:
                            #raise
                            updateUrlsInDatabase(filename_dbs, url, "isignored", 1)
                    except:
                        print("  An error occurred during the access.")
                        updateUrlsInDatabase(filename_dbs, url, "iserror", 1)
                        i = i + 1
                        continue
                else:
                    i = i + 1
                    continue

                # read the HTML content
                try:
                    soup = BeautifulSoup(reqs.text, 'html.parser')
                except:
                    print("  An error occurred during the reading.")
                    updateUrlsInDatabase(filename_dbs, url, "iserror", 1)
                    i = i + 1
                    continue  
                
                # extract the hyperlinks
                urls = []
                for link in soup.find_all('a'):
                    tmp = link.get('href')
                    urls.append(link.get('href'))

                # reconstuct the URL
                # and save the links
                urls2 = []    
                for j in range (0, len(urls)):
                    tmp = urljoin(url, urls[j])

                    o1 = urlparse(tmp)
                    if (o1[5] != ""):
                        tmp = tmp.replace ("#"+o1[5], "") 

                    hyperlink_from.append(url)
                    hyperlink_to.append(tmp)                   

                    urls2.append(tmp)
                    allurls.append(tmp)

                    if (tmp not in wascrawled) and (tmp not in tobecrawled) and (parameter_2=="outside" or (parameter_2=="inside" and getFullDomainFromUrl(tmp) in firstdomains)):
                        tobecrawled.append(tmp)

                addUrlsToDatabase(filename_dbs, urls2)
                del(urls2)
                
                i = i+1
                n = len(tobecrawled)
                if (depth_n[depth_i-1] == i):
                    depth_n.append(n)
                    print ("  Depth level %d is reached." % depth_i)
                    depth_i = depth_i + 1

            addLinksToDatabase(filename_dbs, hyperlink_from, hyperlink_to)

            done = time.time()
            elapsed = done - start
            print ("  The crawl was completed (%s)." % (time.strftime('%Hh %Mmin %Ss', time.gmtime(elapsed))))

    else:
        print ("  A parameter is missing.")
        
    print ("End.")
elif  (parameter_1=="export"):
    print ("Exporting the files...")
    if parameter_2 is not None and parameter_3 is not None:

        filename_urls = parameter_3 + ".txt"
        filename_dbs  = os.path.join(parameter_3, parameter_3 + ".db")
        filename_dot  = os.path.join(parameter_3, parameter_3 + ".dot")
        filename_csv  = os.path.join(parameter_3, parameter_3 + ".csv")        
        go = True
        
        if (os.path.isfile(filename_urls)):
            print ("  The file with URLs exists.")
        else:
            print ("  The file with URLs doesn't exist.")
            go = False
        if (os.path.isfile(filename_dbs)):
            print ("  The file of database exists.")
        else:
            print ("  The file of database doesn't exist.")
            go = False

        if parameter_2 != "graph" and parameter_2 != "csv":
            print ("  The parameter [graph, csv] is not valid.")
            go = False

        if (go==True):

            if (parameter_2 == "graph"):

                start = time.time()
                
                strtmp = "digraph %s_dot {\n" % parameter_3

                con = sqlite3.connect(filename_dbs)
                cur = con.cursor()
                if config['SQL']['sqldot']!="":
                    cur.execute(config['SQL']['sqldot'])
                else:
                    cur.execute("SELECT DISTINCT fulldomain_from, fulldomain_to, COUNT(*) AS occ FROM links WHERE fulldomain_from<>fulldomain_to AND fulldomain_to<>'' GROUP BY hyperlink_from, hyperlink_to HAVING occ>1 ORDER BY occ DESC;")
                rows = cur.fetchall()
                for row in rows:
                    strtmp = strtmp + "\t\"" + row[0] + "\" -> \"" + row[1] + "\";\n"

                strtmp = strtmp + "}"

                fp = open(filename_dot, 'w')
                if fp:
                    fp.write(strtmp)
                    fp.close()

                done = time.time()
                elapsed = done - start

                print ("  The graph file was exported (%.2fs)." % (elapsed))
            elif (parameter_2 == "csv"):

                start = time.time()

                con = sqlite3.connect(filename_dbs)
                cur = con.cursor()
                if config['SQL']['sqlcsv']!="":
                    cur.execute(config['SQL']['sqlcsv'])
                else:
                    cur.execute("SELECT DISTINCT hyperlink_from, hyperlink_to, fulldomain_from, fulldomain_to FROM links WHERE hyperlink_from<>hyperlink_to AND hyperlink_to<>''")
                rows = cur.fetchall()
                strtmp = ""
                for row in rows:
                    strtmp = strtmp + "\"%s\";\"%s\";\"%s\";\"%s\";\"%s\";\"%s\"\n" %(row[0],row[1],row[2],row[3],getDomainNameFromFullDomain(row[2]),getDomainNameFromFullDomain(row[3]))

                fp = open(filename_csv, 'w')
                if fp:
                    fp.write(strtmp)
                    fp.close()

                done = time.time()
                elapsed = done - start
                print ("  The CSV file was exported (%.2fs)." % (elapsed))

    else:
        print ("  A parameter is missing.")
    print ("End.")
else:
    print ("There is no parameter.")

